"""A processor for subbing any datetimes with the dt column."""

import logging

import pandas as pd
from feature_engine.datetime import DatetimeSubtraction

from .identifier import Identifier


def datetimesub_process(
    df: pd.DataFrame, dt_column: str, identifiers: list[Identifier]
) -> pd.DataFrame:
    """Process date time subtractions."""
    identifier_columns = []
    for identifier in identifiers:
        identifier_columns.extend(identifier.columns)
    columns = (
        df.drop(columns=[dt_column])
        .select_dtypes(
            include=["datetime", "datetime64", "datetime64[ns]", "datetimetz"]
        )
        .columns.values.tolist()
    )
    if columns:
        for column in columns:
            if column not in identifier_columns:
                continue
            try:
                dts = DatetimeSubtraction(variables=[column], reference=[dt_column])  # type: ignore
                df = dts.fit_transform(df)
            except TypeError as exc:
                logging.warning(
                    "Couldn't use datetime subtraction on column %s: %s",
                    column,
                    str(exc),
                )
    return df



"""Calculate margin features."""

# pylint: disable=too-many-branches

import pandas as pd
from tqdm import tqdm

from .columns import DELIMITER
from .identifier import Identifier


def margin_process(df: pd.DataFrame, identifiers: list[Identifier]) -> pd.DataFrame:
    """Process margins between teams."""
    tqdm.pandas(desc="Margins Features")
    identifiers_ts: dict[str, dict[str, float]] = {}

    def record_margin(row: pd.Series) -> pd.Series:
        nonlocal identifiers
        nonlocal identifiers_ts

        entity_dicts: dict[str, dict[str, dict[str, float]]] = {}

        for identifier in identifiers:
            if identifier.column not in row:
                continue
            identifier_id = row[identifier.column]
            if pd.isnull(identifier_id):
                continue
            entity_dict = entity_dicts.get(str(identifier.entity_type), {})
            identifier_dict = entity_dict.get(identifier_id, {})
            columns = {
                x[len(identifier.column_prefix) :] for x in identifier.feature_columns
            }
            if identifier.points_column is not None:
                columns.add(identifier.points_column[len(identifier.column_prefix) :])

            for column in columns:
                full_column = identifier.column_prefix + column
                if full_column not in row:
                    continue
                value = row[full_column]
                try:
                    identifier_dict[column] = float(value)
                except TypeError:
                    pass

            entity_dict[identifier_id] = identifier_dict
            entity_dicts[str(identifier.entity_type)] = entity_dict

        for entity_dict in entity_dicts.values():
            max_dict: dict[str, float] = {}
            for identifier_dict in entity_dict.values():
                for column, value in identifier_dict.items():
                    max_dict[column] = max(max_dict.get(column, 0.0), value)
            for identifier_id, identifier_dict in entity_dict.items():
                for identifier in identifiers:
                    if identifier.column not in row:
                        continue
                    identifier_id_check = row[identifier.column]
                    if pd.isnull(identifier_id_check):
                        continue
                    if identifier_id_check != identifier_id:
                        continue
                    identifier_ts_cols: dict[str, float] = identifiers_ts.get(
                        identifier_id, {}
                    )
                    # Copy the old columns in
                    for column, value in identifier_ts_cols.items():
                        full_column = identifier.column_prefix + column
                        row[full_column] = value
                    identifier_ts_cols = {}
                    # Create the new columns
                    for column, value in identifier_dict.items():
                        margin_absolute_column = DELIMITER.join(
                            [column, "margin", "absolute"]
                        )
                        identifier_ts_cols[margin_absolute_column] = (
                            value - max_dict[column]
                        )
                        margin_relative_column = DELIMITER.join(
                            [column, "margin", "relative"]
                        )
                        identifier_ts_cols[margin_relative_column] = (
                            0.0
                            if max_dict[column] == 0.0
                            else (value / max_dict[column])
                        )
                    identifiers_ts[identifier_id] = identifier_ts_cols
                    break

        return row

    return df.progress_apply(record_margin, axis=1)  # type: ignore



"""A description of player representation in the dataframe."""

from .bet import Bet
from .entity_type import EntityType


class Identifier:
    """A way to identify an entity."""

    # pylint: disable=too-many-arguments,too-many-positional-arguments,too-few-public-methods,too-many-instance-attributes

    def __init__(
        self,
        entity_type: EntityType,
        column: str,
        feature_columns: list[str],
        column_prefix: str,
        points_column: str | None = None,
        team_identifier_column: str | None = None,
        field_goals_column: str | None = None,
        assists_column: str | None = None,
        field_goals_attempted_column: str | None = None,
        offensive_rebounds_column: str | None = None,
        turnovers_column: str | None = None,
        bets: list[Bet] | None = None,
        latitude_column: str | None = None,
        longitude_column: str | None = None,
    ):
        self.entity_type = entity_type
        self.column = column
        self.feature_columns = feature_columns
        self.column_prefix = column_prefix
        self.points_column = points_column
        self.team_identifier_column = team_identifier_column
        self.field_goals_column = field_goals_column
        self.assists_column = assists_column
        self.field_goals_attempted_column = field_goals_attempted_column
        self.offensive_rebounds_column = offensive_rebounds_column
        self.turnovers_column = turnovers_column
        self.bets = bets if bets is not None else []
        self.latitude_column = latitude_column
        self.longitude_column = longitude_column

    @property
    def columns(self) -> list[str]:
        """The columns recognised by the identifier."""
        columns = {self.column}
        for feature_column in self.feature_columns:
            columns.add(feature_column)
        if self.points_column is not None:
            columns.add(self.points_column)
        if self.team_identifier_column is not None:
            columns.add(self.team_identifier_column)
        if self.field_goals_column is not None:
            columns.add(self.field_goals_column)
        if self.assists_column is not None:
            columns.add(self.assists_column)
        if self.field_goals_attempted_column is not None:
            columns.add(self.field_goals_attempted_column)
        if self.offensive_rebounds_column is not None:
            columns.add(self.offensive_rebounds_column)
        if self.turnovers_column is not None:
            columns.add(self.turnovers_column)
        if self.latitude_column is not None:
            columns.add(self.latitude_column)
        if self.longitude_column is not None:
            columns.add(self.longitude_column)
        return list(columns)



"""A class describing a team in a match."""

# pylint: disable=too-few-public-methods


class Team:
    """Data about a team in a match."""

    def __init__(self, players: list[str], points: float | None, identifier: str):
        self.players = players
        self.points = points
        self.identifier = identifier



"""Tests for the margin process function."""
import datetime
import unittest

import pandas as pd
from pandas.testing import assert_frame_equal

from sportsfeatures.margin_process import margin_process
from sportsfeatures.identifier import Identifier
from sportsfeatures.entity_type import EntityType


class TestMarginProcess(unittest.TestCase):

    def test_margin_process(self):
        team_0_column_prefix = "teams/0"
        team_1_column_prefix = "teams/1"
        dt_column = "dt"
        team_0_id_column = team_0_column_prefix + "/id"
        team_0_points_column = team_0_column_prefix + "/points"
        team_0_id = "0"
        team_1_id_column = team_1_column_prefix + "/id"
        team_1_points_column = team_1_column_prefix + "/points"
        team_1_id = "1"
        df = pd.DataFrame(data={
            dt_column: [datetime.datetime(2022, 1, 1), datetime.datetime(2022, 1, 2), datetime.datetime(2022, 1, 3)],
            team_0_id_column: [team_0_id, team_1_id, team_0_id],
            team_0_points_column: [10.0, 20.0, 30.0],
            team_1_id_column: [team_1_id, team_0_id, team_1_id],
            team_1_points_column: [20.0, 40.0, 60.0],
        })
        identifiers = [
            Identifier(
                EntityType.TEAM,
                team_0_id_column,
                [],
                team_0_column_prefix,
                points_column=team_0_points_column,
            ),
            Identifier(
                EntityType.TEAM,
                team_1_id_column,
                [],
                team_1_column_prefix,
                points_column=team_1_points_column,
            ),
        ]
        new_df = margin_process(df, identifiers)
        print(new_df)
        print(new_df.columns.values)



"""Process the distance between two locations."""

# pylint: disable=duplicate-code,too-many-branches

import functools

import geopy.distance  # type: ignore
import pandas as pd
from tqdm import tqdm

from .columns import DELIMITER
from .entity_type import EntityType
from .identifier import Identifier


def distance_process(df: pd.DataFrame, identifiers: list[Identifier]) -> pd.DataFrame:
    """Process a dataframe for offensive efficiency."""
    tqdm.pandas(desc="Distance Features")
    last_identifier_locations: dict[str, tuple[float, float]] = {}
    team_identifiers = [x for x in identifiers if x.entity_type == EntityType.TEAM]
    player_identifiers = [x for x in identifiers if x.entity_type == EntityType.PLAYER]
    venue_identifiers = [x for x in identifiers if x.entity_type == EntityType.VENUE]

    def record_distance(
        row: pd.Series,
        team_identifiers: list[Identifier],
        player_identifiers: list[Identifier],
        venue_identifiers: list[Identifier],
    ) -> pd.Series:
        nonlocal identifiers
        nonlocal last_identifier_locations

        current_location = None
        for venue_identifier in venue_identifiers:
            if venue_identifier.latitude_column is None:
                continue
            if venue_identifier.latitude_column not in row:
                continue
            latitude = row[venue_identifier.latitude_column]
            if pd.isnull(latitude):
                continue
            if venue_identifier.longitude_column is None:
                continue
            if venue_identifier.longitude_column not in row:
                continue
            longitude = row[venue_identifier.longitude_column]
            if pd.isnull(longitude):
                continue
            current_location = (latitude, longitude)
        if current_location is None:
            return row

        for identifier in team_identifiers + player_identifiers:
            if identifier.column not in row:
                continue
            identifier_id = row[identifier.column]
            if pd.isnull(identifier_id):
                continue
            if not isinstance(identifier_id, str):
                continue
            key = "_".join([str(identifier.entity_type), identifier_id])
            last_location = last_identifier_locations.get(key)
            if last_location is not None:
                row[DELIMITER.join([identifier.column_prefix, "latitudediff"])] = abs(
                    last_location[0] - current_location[0]
                )
                row[DELIMITER.join([identifier.column_prefix, "longitudediff"])] = abs(
                    last_location[1] - current_location[1]
                )
                row[DELIMITER.join([identifier.column_prefix, "distance"])] = (
                    geopy.distance.geodesic(last_location, current_location).km
                )
            last_identifier_locations[key] = current_location

        return row

    return df.progress_apply(
        functools.partial(
            record_distance,
            team_identifiers=team_identifiers,
            player_identifiers=player_identifiers,
            venue_identifiers=venue_identifiers,
        ),
        axis=1,
    )  # type: ignore



"""A description of a bet in the dataframe."""


class Bet:
    """A way to define a bet."""

    # pylint: disable=too-few-public-methods

    def __init__(
        self,
        odds_column: str,
        bookie_id_column: str,
        dt_column: str | None = None,
    ):
        self.odds_column = odds_column
        self.bookie_id_column = bookie_id_column
        self.dt_column = dt_column



"""An enum containing the different entity types."""

from enum import StrEnum, auto


class EntityType(StrEnum):
    """The type of entity being represented."""

    PLAYER = auto()
    TEAM = auto()
    VENUE = auto()



"""Tests for the datetimesub process function."""
import datetime
import unittest

import pandas as pd

from sportsfeatures.datetimesub_process import datetimesub_process
from sportsfeatures.identifier import Identifier
from sportsfeatures.entity_type import EntityType


class TestDatetimesubProcess(unittest.TestCase):

    def test_datetimesub_process(self):
        dt_other_column = "dt_other"
        identifier = Identifier(EntityType.TEAM, "team/0/id", [dt_other_column], "team/0/")
        dt_column = "dt"
        df = pd.DataFrame(data={
            dt_column: [datetime.datetime(2022, 1, 1), datetime.datetime(2022, 1, 2), datetime.datetime(2022, 1, 3)],
            dt_other_column: [datetime.datetime(2022, 1, 1, tzinfo=datetime.timezone.utc), datetime.datetime(2022, 1, 2, tzinfo=datetime.timezone.utc), datetime.datetime(2022, 1, 3, tzinfo=datetime.timezone.utc)],
        })
        datetimesub_process(df, dt_column, [identifier])



"""Calculate bet features."""

import datetime
import functools

import pandas as pd
from sklearn.metrics import mean_squared_error  # type: ignore
from tqdm import tqdm

from .columns import DELIMITER
from .identifier import Identifier


def _force_utc_aware(series, timezone="UTC"):
    series = pd.to_datetime(series, errors="coerce")

    if series.dt.tz is None:
        return series.dt.tz_localize(timezone, ambiguous="NaT", nonexistent="NaT")
    return series.dt.tz_convert(timezone)


def bet_process(
    df: pd.DataFrame, identifiers: list[Identifier], dt_column: str
) -> pd.DataFrame:
    """Process bets."""
    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    tqdm.pandas(desc="Bets Features")
    bookie_odds: list[float] = []
    wins: list[float] = []

    def apply_bets(
        row: pd.Series, identifiers: list[Identifier], dt_column: str
    ) -> pd.Series:
        nonlocal bookie_odds
        nonlocal wins

        try:
            game_dt = pd.Timestamp(row[dt_column]).tz_localize("UTC")
        except TypeError:
            game_dt = pd.Timestamp(row[dt_column]).tz_convert("UTC")

        price_efficiency = (
            None if not bookie_odds else mean_squared_error(bookie_odds, wins)
        )
        local_bookie_odds = []
        local_points = []
        for identifier in identifiers:
            if identifier.points_column is None:
                continue
            if identifier.points_column not in row:
                continue
            points = row[identifier.points_column]
            if pd.isnull(points):
                continue

            odds_data = []
            bookies_data = []
            dts_data = []
            for bet in identifier.bets:
                if bet.odds_column not in row or bet.bookie_id_column not in row:
                    continue
                odds = row[bet.odds_column]
                if pd.isnull(odds):
                    continue
                bookie_id = row[bet.bookie_id_column]
                if pd.isnull(bookie_id):
                    continue
                if bet.dt_column is None or bet.dt_column not in row:
                    dt = game_dt - datetime.timedelta(hours=1)
                else:
                    dt = row[bet.dt_column]
                    if pd.isnull(dt):
                        dt = game_dt - datetime.timedelta(hours=1)
                odds_data.append(odds)
                bookies_data.append(bookie_id)
                dts_data.append(dt)
            df = pd.DataFrame(
                data={
                    "odds": odds_data,
                    "bookie": bookies_data,
                    "dt": dts_data,
                }
            )
            if df.empty:
                continue
            df["dt"] = _force_utc_aware(df["dt"])
            df = df.sort_values(by="dt")
            odds_max = df["odds"].max()
            odds_min = df["odds"].min()
            earliest_odds = df["odds"].iloc[0]
            latest_odds = df["odds"].iloc[-1]
            earliest_dt = df["dt"].iloc[0]
            latest_dt = df["dt"].iloc[-1]

            direction_changes = 0
            big_shifts = 0
            consensus_flips = 0
            resampled_df = pd.DataFrame()
            for bookie in df["bookie"].unique():
                bookie_df = df[df["bookie"] == bookie]
                bookies_odds = bookie_df["odds"].to_list()
                current_odds = None
                current_direction = None
                for odd in bookies_odds:
                    if current_odds is None:
                        current_odds = odd
                        continue
                    if current_direction is None:
                        current_direction = 1 if odd >= current_odds else -1
                        current_odds = odd
                        continue
                    new_direction = 1 if odd >= current_odds else -1
                    if new_direction != current_direction:
                        direction_changes += 1
                    if max(current_odds, odd) / min(current_odds, odd) > 1.1:
                        big_shifts += 1
                    if min(current_odds, odd) < 2.0 < max(current_odds, odd):
                        consensus_flips += 1
                    current_odds = odd
                    current_direction = new_direction
                resampled_df = pd.concat(
                    [
                        resampled_df,
                        pd.DataFrame(
                            data={
                                bookie + "_odds": bookies_odds,
                            },
                            index=bookie_df["dt"],
                        ),
                    ]
                )
            resampled_df = resampled_df.resample("5T").last()
            resampled_df = resampled_df.ffill()

            ffill_df = df.set_index("dt").ffill()
            oneday_df = ffill_df[ffill_df.index > game_dt - datetime.timedelta(days=1)]
            final_odds = resampled_df.mean(axis=1).to_list()[-1]

            row[DELIMITER.join([identifier.column_prefix, "odds", "max"])] = odds_max
            row[DELIMITER.join([identifier.column_prefix, "odds", "min"])] = odds_min
            row[DELIMITER.join([identifier.column_prefix, "odds", "mean"])] = df[
                "odds"
            ].mean()
            row[DELIMITER.join([identifier.column_prefix, "odds", "median"])] = df[
                "odds"
            ].median()
            row[DELIMITER.join([identifier.column_prefix, "odds", "spread"])] = (
                odds_max - odds_min
            )
            row[DELIMITER.join([identifier.column_prefix, "odds", "bookies"])] = df[
                "bookie"
            ].nunique()
            row[DELIMITER.join([identifier.column_prefix, "odds", "roc"])] = (
                latest_odds - earliest_odds
            ) / (latest_dt - earliest_dt).total_seconds()
            row[DELIMITER.join([identifier.column_prefix, "odds", "mom"])] = (
                latest_odds - earliest_odds
            )
            row[DELIMITER.join([identifier.column_prefix, "odds", "directchanges"])] = (
                direction_changes
            )
            row[DELIMITER.join([identifier.column_prefix, "odds", "samples"])] = len(df)
            row[DELIMITER.join([identifier.column_prefix, "odds", "ewm"])] = (
                resampled_df.mean(axis=1).ewm(alpha=0.2, adjust=False).mean()
            )
            row[DELIMITER.join([identifier.column_prefix, "odds", "bigshifts"])] = (
                big_shifts
            )
            row[
                DELIMITER.join([identifier.column_prefix, "odds", "consensusflips"])
            ] = consensus_flips
            row[DELIMITER.join([identifier.column_prefix, "odds", "roc1day"])] = (
                oneday_df["odds"].iloc[-1] - oneday_df["odds"].iloc[0]
            ) / datetime.timedelta(days=1).total_seconds()
            row[DELIMITER.join([identifier.column_prefix, "odds"])] = final_odds
            row[
                DELIMITER.join([identifier.column_prefix, "odds", "priceefficiency"])
            ] = price_efficiency
            local_bookie_odds.append(1.0 / final_odds)
            local_points.append(points)

        bookie_odds.extend(local_bookie_odds)
        wins.extend([float(x == max(local_points)) for x in local_points])

        return row

    return df.progress_apply(
        functools.partial(
            apply_bets,
            identifiers=identifiers,
            dt_column=dt_column,
        ),
        axis=1,
    )  # type: ignore
